import os
import data
import pygame
import random
import math
import pickle

# globals
screen_size = ( 800, 600 )
caption = "CQ Robot Day 7"
background_color = ( 0, 0, 0 )
highlight_selected_tile_color = ( 0, 255, 0 )
catch_area = ( (500,275), (250,300) ) # left upper x/y and width/height
maze_play_area = ( (25,175), (400,400) )    # left upper x/y and width/height
maze_edit_area = ( (75,75), (450,500) )    # left upper x/y and width/height
maze_dimensions = ( 12, 12 )
maze_tiles_area = ( (10,10), (700,50) ) # left upper x/y and width/height
gfx = {}
ball_speed = 2
new_ball_probability = 10
ticks_per_execution = 10
fixed_mouse_pos = (screen_size[0]/2,screen_size[1]/2)
max_nof_balls_in_progress_chain = 6
max_nof_balls_in_robot_chain = 15
frames_per_second = 30.0
result_screen_dimensions = ( 450,350 )
result_screen_text_color = ( 0, 0, 0 )

progress_chain_center_left = ( 500,200 )
robot_chain_centerleft = ( 50,125 )
DEFAULT_LEVEL_FILENAME = "default.lvl"

# tile indeces
ID_CMD_BALL_LEFT     = 0
ID_CMD_BALL_BACKWARD = 1
ID_CMD_BALL_FORWARD  = 2
ID_CMD_BALL_RIGHT    = 3
ID_CMD_BALL_FINISH   = 4
ID_CMD_BALL_UNDO     = 5
ID_CATCHER           = 6
ID_BUTTON            = 7
ID_NOF_CATCH_TILES   = 8
ID_NOF_BALLS         = 6

ID_MAZE_ROBOT         = 0
ID_MAZE_FLOOR_BGRIGHT = 1
ID_MAZE_FLOOR_DARK    = 2
ID_MAZE_VWALL     = 3
ID_MAZE_CORNER_WN = 4
ID_MAZE_CORNER_NE = 5
ID_MAZE_CORNER_ES = 6
ID_MAZE_CORNER_SW = 7
ID_MAZE_EXIT      = 8
ID_MAZE_SHADED    = 9
ID_NOF_MAZE_TILES = 10

ROBOT_STATUS_WAITING = 0
ROBOT_STATUS_EXECUTE = 1

# gfx keys
GFX_KEY_MASK = "game_mask"
GFX_KEY_CATCH_TILES = "catch_tiles"
GFX_KEY_MAZE_TILES = "maze_tiles"
GFX_KEY_MAIN_MENU = "main_menu"
GFX_KEY_BACKGROUND = "background"
GFX_KEY_RESULT = "result"

LEVEL_FILES = [ "level1.dat", "level2.dat", "level3.dat", "level4.dat", "level5.dat", "level6.dat" ]

CATCH_TILES_TEXTURE_WIDTH  = 1600
CATCH_TILES_TEXTURE_HEIGHT = 100
CATCH_TILE_WIDTH = 100
CATCH_TILE_HEIGHT = 100

MAZE_TILE_WIDTH = 50
MAZE_TILE_HEIGHT = 50
MAZE_TEXTURE_WIDTH = 55
MAZE_TEXTURE_HEIGHT = 55

COMMAND_BALL_AREA_WIDTH = 50
COMMAND_BALL_AREA_HEIGHT = 50

CATCHER_WIDTH = 25
CATCHER_HEIGHT = 10

GAME_PAUSED_TEXT = "press <SPACE> to continue"
GAME_PAUSED_COLOR = ( 255, 0, 0 )
GAME_BUTTON_TEXT_COLOR = ( 0, 0, 0 )
GAME_BUTTON_TEXT_COLOR_DISABLED = ( 122, 122, 122 )

ID_NOF_1_COMBOS = 1
ID_NOF_2_COMBOS = 2
ID_NOF_3_COMBOS = 3
ID_NOF_4_COMBOS = 4
ID_NOF_5_COMBOS = 5
ID_NOF_6_COMBOS = 6
ID_NOF_POINTS   = 7
ID_FINISH_REASON = 8

GAME_MODE_MAIN_MENU = "MENU"
GAME_MODE_PLAY_LEVEL = "PLAY"
GAME_MODE_HELP_SCREEN = "HELP"
GAME_MODE_LEVEL_EDITOR = "EDITOR"
GAME_MODE_RESULT_SCREEN = "RESULT"

def is_click_inside_area( click_pos, area ):
  # check if click lies inside the given area
  if click_pos[0] < area[0][0]:
    return False
  if click_pos[0] > area[0][0]+area[1][0]:
    return False
  if click_pos[1] < area[0][1]:
    return False
  if click_pos[1] > area[0][1]+area[1][1]:
    return False
  return True

def draw_text_center( text, pos, color, font, screen ):
  surf = font.render( text, 1, color )
  surf_pos = surf.get_rect()
  surf_pos.center = pos
  screen.blit( surf, surf_pos )
  return

class Button:
  def __init__( self, function, dimensions, button_surface, text, font, color=GAME_BUTTON_TEXT_COLOR ):
    self.dimensions = dimensions # ( (x,y), (width,height) )
    self.text = text
    self.font = font
    self.center = ( dimensions[0][0]+dimensions[1][0]/2, dimensions[0][1]+dimensions[1][1]/2 )
    self.color = color
    self.button_surface = button_surface
    self.function = function
    self.enabled = True
    return
  
  def draw( self, screen ):
    # draw button
    screen.blit( self.button_surface, self.dimensions[0] )
    # draw text
    if self.enabled:
      draw_text_center( self.text , self.center, self.color, self.font, screen)
    else:
      draw_text_center( self.text , self.center, GAME_BUTTON_TEXT_COLOR_DISABLED, self.font, screen)
    return
  
  def handle_click_event( self, pos ):
    if self.enabled == False:
      return
    if is_click_inside_area(pos,self.dimensions):
      self.function()
    return

class Level:
  def __init__( self ):
    self.maze = []
    self.maze.append( [] )
    self.robot_start = ( -1, -1 )
    self.robot_exit  = ( -1, -1 )
    self.time_limit = 60.0
    self.point_limit = 30
    self.max_points_earned = 0
    self.hints = [ "Hint 1", "Hint 2", "Hint 3" ]
    self.ball_speed = 1.0
    self.nof_balls = 5
    return

class Maze:
  def __init__( self, containers ):
    self.containers = containers
    self.level = Level()
    self.robot_position = self.level.robot_start
    self.robot_status = ROBOT_STATUS_WAITING
    self.robot_path = []
    self.execution_ticks_left = 0
    self.current_command_ball = -1
    self.init_resources()
    self.init_test_level()
    self.time_left = 0.0
    return
  
  def init_resources( self ):
    self.walkable = {}
    self.walkable[ID_MAZE_ROBOT] = False
    self.walkable[ID_MAZE_FLOOR_BGRIGHT] = True
    self.walkable[ID_MAZE_FLOOR_DARK] = True
    self.walkable[ID_MAZE_VWALL] = False
    self.walkable[ID_MAZE_EXIT]  = True
    self.walkable[ID_MAZE_CORNER_WN] = False
    self.walkable[ID_MAZE_CORNER_NE] = False
    self.walkable[ID_MAZE_CORNER_ES] = False
    self.walkable[ID_MAZE_CORNER_SW] = False
    self.tile_width = maze_play_area[1][0] / maze_dimensions[0]
    self.tile_height = maze_play_area[1][1] / maze_dimensions[1]
    self.surfaces = {}
    self.surfaces[ID_MAZE_ROBOT] = get_scaled_tile( GFX_KEY_MAZE_TILES , ID_MAZE_ROBOT, self.tile_width, self.tile_height)
    self.surfaces[ID_MAZE_FLOOR_BGRIGHT] = get_scaled_tile( GFX_KEY_MAZE_TILES , ID_MAZE_FLOOR_BGRIGHT, self.tile_width, self.tile_height)
    self.surfaces[ID_MAZE_FLOOR_DARK] = get_scaled_tile( GFX_KEY_MAZE_TILES , ID_MAZE_FLOOR_DARK, self.tile_width, self.tile_height)
    self.surfaces[ID_MAZE_VWALL] = get_scaled_tile( GFX_KEY_MAZE_TILES , ID_MAZE_VWALL, self.tile_width, self.tile_height)
    self.surfaces[ID_MAZE_CORNER_WN] = get_scaled_tile( GFX_KEY_MAZE_TILES , ID_MAZE_CORNER_WN, self.tile_width, self.tile_height)
    self.surfaces[ID_MAZE_CORNER_NE] = get_scaled_tile( GFX_KEY_MAZE_TILES , ID_MAZE_CORNER_NE, self.tile_width, self.tile_height)
    self.surfaces[ID_MAZE_CORNER_ES] = get_scaled_tile( GFX_KEY_MAZE_TILES , ID_MAZE_CORNER_ES, self.tile_width, self.tile_height)
    self.surfaces[ID_MAZE_CORNER_SW] = get_scaled_tile( GFX_KEY_MAZE_TILES , ID_MAZE_CORNER_SW, self.tile_width, self.tile_height)
    self.surfaces[ID_MAZE_SHADED]    = get_scaled_tile( GFX_KEY_MAZE_TILES , ID_MAZE_SHADED,  self.tile_width, self.tile_height)
    self.surfaces[ID_MAZE_EXIT]  = get_scaled_tile( GFX_KEY_MAZE_TILES , ID_MAZE_EXIT,  self.tile_width, self.tile_height)
    return
  
  def reset_level( self, tile=ID_MAZE_FLOOR_BGRIGHT ):
    self.level = Level()
    for x in range(maze_dimensions[0]):
      self.level.maze.append( [] )
      for y in range(maze_dimensions[1]):
        self.level.maze[x].append( tile )
    self.containers.reset_all_containers()
    self.robot_position = (-1,-1)
    self.robot_status = ROBOT_STATUS_WAITING
    self.robot_path = []
    self.execution_ticks_left = 0
    self.current_command_ball = -1
    self.containers.robot_execution_index = -1
    self.time_left = 0.0
    return
  
  def save_level( self, filename=DEFAULT_LEVEL_FILENAME ):
    f = open( DEFAULT_LEVEL_FILENAME, "w" )
    pickle.dump( self.level, f )
    f.close()
    return
  
  def load_level( self, filename=DEFAULT_LEVEL_FILENAME ):
    f = open( filename, "r" )
    self.level = pickle.load( f )
    f.close()
    self.robot_position = self.level.robot_start
    self.time_left = self.level.time_limit
    self.level.max_points_earned = 0
    return
  
  def handle_click_event( self, mouse_pos, tile_id ):
    # check if click lies inside the maze area
    if mouse_pos[0] < maze_edit_area[0][0]:
      return
    if mouse_pos[0] > maze_edit_area[0][0]+maze_dimensions[0]*self.tile_width:
      return
    if mouse_pos[1] < maze_edit_area[0][1]:
      return
    if mouse_pos[1] > maze_edit_area[0][0]+maze_dimensions[1]*self.tile_height:
      return
    
    # calculate clicked index
    x = int( (mouse_pos[0]-maze_edit_area[0][0])/self.tile_width )
    y = int( (mouse_pos[1]-maze_edit_area[0][1])/self.tile_height )
    
    if tile_id == ID_MAZE_ROBOT:
      self.level.robot_start = (x,y)
      return
    
    if tile_id == ID_MAZE_EXIT:
      if self.level.robot_exit != (-1,-1):
        self.level.maze[self.level.robot_exit[0]][self.level.robot_exit[1]] = ID_MAZE_FLOOR_BGRIGHT
      self.level.robot_exit = (x,y)
      self.level.maze[self.level.robot_exit[0]][self.level.robot_exit[1]] = tile_id
      return
    
    # set tile
    self.level.maze[x][y] = tile_id
    
    return
  
  def init_test_level( self ):
    self.level.maze = []
    self.level.maze.append( [1,1,1,1,2,1,1,1,1,2,1,1,1,1,2,1,1,1,1,2] )
    self.level.maze.append( [1,2,1,1,1,1,1,1,1,2,1,2,1,1,1,1,1,1,1,2] )
    self.level.maze.append( [1,2,3,3,1,3,3,3,3,1,1,2,3,3,1,3,3,3,3,2] )
    self.level.maze.append( [3,2,1,1,1,1,1,1,1,2,3,2,1,1,1,1,1,1,1,2] )
    self.level.maze.append( [3,3,3,3,3,3,3,3,1,2,3,3,3,3,3,3,3,3,1,2] )
    self.level.maze.append( [1,1,1,1,1,1,1,1,1,2,1,1,1,1,1,1,1,1,1,2] )
    self.level.maze.append( [1,1,1,1,2,3,3,3,3,1,1,1,1,1,2,3,3,3,3,2] )
    self.level.maze.append( [1,1,1,1,1,1,1,1,1,2,1,1,1,1,1,1,1,1,1,2] )
    self.level.maze.append( [1,1,1,1,2,1,1,1,1,2,1,1,1,1,2,1,1,1,1,2] )
    self.level.maze.append( [1,1,1,1,2,1,1,1,1,2,1,1,1,1,2,1,1,1,1,2] )
    self.level.maze.append( [1,1,1,1,2,1,1,1,1,1,1,1,1,1,2,1,1,1,1,2] )
    self.level.maze.append( [1,2,1,1,1,1,1,1,1,2,1,2,1,1,1,1,1,1,1,2] )
    self.level.maze.append( [1,2,3,3,1,3,3,3,3,2,1,2,3,3,1,3,3,3,3,2] )
    self.level.maze.append( [3,2,1,1,1,1,1,1,1,2,3,2,1,1,1,1,1,1,1,2] )
    self.level.maze.append( [3,3,3,3,3,3,3,3,1,2,3,3,3,3,3,3,3,3,1,2] )
    self.level.maze.append( [1,1,1,1,1,1,1,1,1,2,1,1,1,1,1,1,1,1,1,2] )
    self.level.maze.append( [1,1,1,1,2,3,3,3,3,2,1,1,1,1,2,3,3,3,3,2] )
    self.level.maze.append( [1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,2] )
    self.level.maze.append( [1,1,1,1,2,1,1,1,1,2,1,1,1,1,2,1,1,1,1,2] )
    self.level.maze.append( [1,1,1,1,2,1,1,1,1,2,1,1,1,1,2,1,1,1,1,2] )
    # place robot
    rand_x = random.randrange(self.level.maze.__len__())
    rand_y = random.randrange(self.level.maze.__len__())
    self.level.maze[rand_x][rand_y] = ID_MAZE_EXIT
    
    rand_x = random.randrange(self.level.maze.__len__())
    rand_y = random.randrange(self.level.maze.__len__())
    self.robot_position = ( rand_x,rand_y )
    self.robot_status = ROBOT_STATUS_WAITING
    self.execution_ticks_left = 0
    self.current_command_ball = None
    return
  
  def execute_command( self, type, checkShade=True ):
    new_position = ( 0, 0 )
    if type == ID_CMD_BALL_FORWARD:
      new_position = ( self.robot_position[0], self.robot_position[1]-1 )
    
    if type == ID_CMD_BALL_LEFT:
      new_position = ( self.robot_position[0]-1, self.robot_position[1] )
    
    if type == ID_CMD_BALL_RIGHT:
      new_position = ( self.robot_position[0]+1, self.robot_position[1] )
    
    if type == ID_CMD_BALL_BACKWARD:
      new_position = ( self.robot_position[0], self.robot_position[1]+1 )
    
    # check bond of new position
    if new_position[0]<0 or new_position[0]>=maze_dimensions[0]:
      return False
    if new_position[1]<0 or new_position[1]>=maze_dimensions[1]:
      return False
    
    # check if the robot was here before
    if checkShade:
      try:
        self.robot_path.index( new_position )
        return False
      except:
        pass
    
    tile_at_new_position = self.level.maze[new_position[0]][new_position[1]]
    if self.walkable[tile_at_new_position]:
      self.robot_position = new_position
      return True
    else:
      return False
  
  def robot_tick( self, containers ):
    self.time_left = self.time_left - 1/frames_per_second
    
    if self.robot_position == (-1,-1):
      return
    
    if self.containers.undo != None:
      self.execute_command( self.containers.undo.getOppositeType(), False )
      self.containers.undo = None
      self.robot_path.pop()
    
    if self.robot_status == ROBOT_STATUS_WAITING:
      if self.containers.robot_execution_index != -1:
        # get the next command ball
        while self.containers.robotq[self.containers.robot_execution_index].type == ID_CMD_BALL_FINISH:
          self.containers.robot_execution_index = self.containers.robot_execution_index + 1
          if self.containers.robot_execution_index >= self.containers.robotq.__len__()-1:
            self.containers.robot_execution_index = -1
            return
        ball = self.containers.robotq[self.containers.robot_execution_index]
        
        # execute this ball
        previous_position = self.robot_position
        if self.execute_command( ball.type ):
          self.robot_path.append( previous_position )
          # increase index
          self.containers.robot_execution_index = self.containers.robot_execution_index + 1
        else:
          # the path is not clear so abort the execution and remove all following commands
          while self.containers.robotq.__len__()>self.containers.robot_execution_index+1:
            # remove the next to the last. the last is a finish ball
            self.containers.balls.remove( self.containers.robotq.pop(-2) )
          if self.containers.robotq.__len__() == 1:
            self.containers.balls.remove( self.containers.robotq.pop() )
          else:
            if self.containers.robotq[-1].type == self.containers.robotq[-2].type:
              # remove adjacent finish ball
              self.containers.balls.remove( self.containers.robotq.pop() )
          self.containers.robot_execution_index = -1
          self.containers.update_ball_positions()
    return
  
  def draw( self, screen, mode=GAME_MODE_PLAY_LEVEL ):
    start_x = maze_play_area[0][0]
    start_y = maze_play_area[0][1]
    
    if mode == GAME_MODE_LEVEL_EDITOR:
      start_x = maze_edit_area[0][0]
      start_y = maze_edit_area[0][1]
    
    step_x = self.tile_width
    step_y = self.tile_height
    # draw maze
    for x in range(maze_dimensions[0]):
      for y in range(maze_dimensions[1]):
        tile_id = self.level.maze[x][y]
        screen.blit( self.surfaces[tile_id], (start_x+x*step_x,start_y+y*step_y) )
    
    # draw path shades
    if mode == GAME_MODE_PLAY_LEVEL:
      for point in self.robot_path:
        x = point[0]
        y = point[1]
        screen.blit( self.surfaces[ID_MAZE_SHADED], (start_x+x*step_x,start_y+y*step_y) )
    
    # draw robot
    if mode == GAME_MODE_PLAY_LEVEL:
      screen.blit( self.surfaces[ID_MAZE_ROBOT], (start_x+self.robot_position[0]*step_x,start_y+self.robot_position[1]*step_y) )
    elif mode == GAME_MODE_LEVEL_EDITOR:
      screen.blit( self.surfaces[ID_MAZE_ROBOT], (start_x+self.level.robot_start[0]*step_x,start_y+self.level.robot_start[1]*step_y) )
      
    return
    

class Containers:
  def __init__( self ):
    self.balls = pygame.sprite.RenderPlain()
    self.cq = []
    self.robotq = []
    self.max_cq_size = 6
    self.robot_execution_index = -1
    self.cq_centerleft = progress_chain_center_left
    self.robotq_centerleft = robot_chain_centerleft
    self.undo = None
    return
  
  def reset_all_containers( self ):
    self.balls = pygame.sprite.RenderPlain()
    self.cq = []
    self.robotq = []
    self.undo = None
    return
  
  def add_ball( self, ball ):
    if ball.type == ID_CMD_BALL_FORWARD:
      self.add_to_cq( ball )
    if ball.type == ID_CMD_BALL_LEFT:
      self.add_to_cq( ball )
    if ball.type == ID_CMD_BALL_RIGHT:
      self.add_to_cq( ball )
    if ball.type == ID_CMD_BALL_BACKWARD:
      self.add_to_cq( ball )
    if ball.type == ID_CMD_BALL_FINISH:
      self.transfer_balls_to_robot( ball )
    if ball.type == ID_CMD_BALL_UNDO:
      self.erase_last_ball_from_robotq()
    return
  
  def erase_last_ball_from_robotq( self ):
    if self.robotq.__len__()==0:
      return
    
    # get the ball to erase
    self.undo = self.robotq.pop(-2)
    self.balls.remove( self.undo )
    
    # remove the finish ball of a possible empty command tokens
    if self.robotq.__len__()>1 and self.robotq[-2].type==ID_CMD_BALL_FINISH:
      ball = self.robotq.pop(-2)
      self.balls.remove( ball )
    
    # if there is only one ball in the queue it must be a finish ball so delete it
    if self.robotq.__len__()==1:
      ball = self.robotq.pop()
      self.balls.remove( ball )
    
    self.update_ball_positions()
    return
  
  def transfer_balls_to_robot( self, finish ):
    if self.cq.__len__() == 0:
      return
    
    # if the robot isn't working, give him the start of the new balls 
    if self.robot_execution_index == -1:
      self.robot_execution_index = self.robotq.__len__() 
    
    while self.cq.__len__() > 0:
      ball = self.cq.pop(0)
      self.robotq.append( ball )
    
    self.robotq.append( finish )
    self.balls.add( finish )
    
    self.update_ball_positions()
    return
  
  def update_ball_positions( self ):
    # command queue 
    i = 0
    for ball in self.cq:
      ball.rect.center = ( self.cq_centerleft[0]+COMMAND_BALL_AREA_WIDTH*i,self.cq_centerleft[1] )
      i = i+1
    
    # robot queue
    robotqlen = self.robotq.__len__()
    max_nof_balls = max_nof_balls_in_robot_chain
    if self.robotq.__len__() > max_nof_balls_in_robot_chain:
      # set the front balls, that have no space in the queue, outside
      for i in range(self.robotq.__len__()-max_nof_balls_in_robot_chain):
        ball = self.robotq[i]
        ball.rect.center = ( 0,0 )
      # place the rest of the balls in the robot queue to the screen
      for i in range(max_nof_balls_in_robot_chain):
        index = i + self.robotq.__len__()-max_nof_balls_in_robot_chain
        ball = self.robotq[index]
        ball.rect.center = ( self.robotq_centerleft[0]+COMMAND_BALL_AREA_WIDTH*i,self.robotq_centerleft[1] )
    else:
      for ball in self.robotq:
        index = self.robotq.index(ball)
        ball.rect.center = ( self.robotq_centerleft[0]+COMMAND_BALL_AREA_WIDTH*index,self.robotq_centerleft[1] )
    
    return
  
  def add_to_cq( self, ball ):
    # remove opponent directions
    if self.cq.__len__() > 0:
      last = self.cq.pop()
      if ball.type==ID_CMD_BALL_FORWARD and last.type==ID_CMD_BALL_BACKWARD:
        self.balls.remove( last )
        return
      if ball.type==ID_CMD_BALL_BACKWARD and last.type==ID_CMD_BALL_FORWARD:
        self.balls.remove( last )
        return
      if ball.type==ID_CMD_BALL_LEFT and last.type==ID_CMD_BALL_RIGHT:
        self.balls.remove( last )
        return
      if ball.type==ID_CMD_BALL_RIGHT and last.type==ID_CMD_BALL_LEFT:
        self.balls.remove( last )
        return
      self.cq.append( last )
    
    if self.cq.__len__() < self.max_cq_size:
      self.cq.append( ball )
      self.balls.add( ball )
    
    self.update_ball_positions()
    return
  
  def get_command_ball_from_cq( self ):
    if self.cq.__len__() > 0:
      ball = self.cq.pop( 0 )
      self.balls.remove( ball )
      return ball 
    else:
      return None
  
class Catcher( pygame.sprite.Sprite ):
  def __init__( self, position, sprite_group ):
    pygame.sprite.Sprite.__init__( self )
    self.image = get_scaled_tile( GFX_KEY_CATCH_TILES ,ID_CATCHER, CATCHER_WIDTH, CATCHER_HEIGHT)
    self.rect = self.image.get_rect() 
    self.rect.center = position
    self.sprite_group = sprite_group
    self.sprite_group.add( self )

class CommandBall( pygame.sprite.Sprite ):
  def __init__( self, position, sprite_group, type=ID_CMD_BALL_FORWARD ):
    pygame.sprite.Sprite.__init__( self )
    self.type = type
    self.image = get_scaled_tile( GFX_KEY_CATCH_TILES, type, COMMAND_BALL_AREA_WIDTH, COMMAND_BALL_AREA_HEIGHT )
    self.rect = self.image.get_rect() 
    self.rect.center = position
    self.sprite_group = sprite_group
    self.sprite_group.add( self )
    return
  
  def getOppositeType(self, type=None):
    if type == None:
      type = self.type
    if type == ID_CMD_BALL_BACKWARD:
      return ID_CMD_BALL_FORWARD
    if type == ID_CMD_BALL_FORWARD:
      return ID_CMD_BALL_BACKWARD
    if type == ID_CMD_BALL_LEFT:
      return ID_CMD_BALL_RIGHT
    if type == ID_CMD_BALL_RIGHT:
      return ID_CMD_BALL_LEFT
  
  def tick( self, speed ):
    self.rect = self.rect.move( 0, speed )
    if self.rect.top > catch_area[0][1]+catch_area[1][1]:
      self.sprite_group.remove( self )
    return
    
def get_scaled_tile( key, index, width, height ):
    surface_rect = ( index*CATCH_TILE_WIDTH, 0, CATCH_TILE_WIDTH-1, CATCH_TILE_HEIGHT-1 )
    return pygame.transform.scale( gfx[key].subsurface(surface_rect), (width,height))

def load_image( key, file, new_width=0, new_height=0 ):
  # create filename
  filename = os.path.join( 'data', file )
  try:
      surface = pygame.image.load( filename )
  except pygame.error:
      raise SystemExit, 'Could not load image "%s" %s'%( filename, pygame.get_error())
  gfx[key] = surface.convert_alpha()
  if new_width>0 and new_height>0:
    gfx[key] = pygame.transform.scale( gfx[key], (new_width,new_height) )
  return

def get_distance( p1, p2 ):
  xDiff = p1[0]-p2[0]
  yDiff = p1[1]-p2[1]
  return math.sqrt( xDiff*xDiff + yDiff*yDiff )

def get_new_command_ball_position( ball_group ):
  x = catch_area[0][0] + COMMAND_BALL_AREA_WIDTH/2 + random.randrange(catch_area[1][0]-COMMAND_BALL_AREA_WIDTH/2)
  y = catch_area[0][1] - COMMAND_BALL_AREA_HEIGHT/2
  # determine if the position conflicts with an existing command ball
  for ball in ball_group:
    distance = get_distance( ball.rect.center,(x,y) )
    if distance < COMMAND_BALL_AREA_HEIGHT:
      return (0,0)
  return (x,y)


class Game:
  def __init__( self ):
    pygame.init()
    pygame.display.init()
    self.screen = pygame.display.set_mode( screen_size, pygame.FULLSCREEN )
    #self.screen = pygame.display.set_mode( screen_size )
    pygame.display.set_caption( caption )
    self.large_font = pygame.font.Font( None, 40 )
    self.small_font = pygame.font.Font( None, 30 )
    self.load_resources()
    self.run_game = True
    self.game_paused = False
    self.containers = Containers()
    self.maze = Maze( self.containers )
    self.ball_group = pygame.sprite.RenderPlain()
    self.catcher_group = pygame.sprite.RenderPlain()
    self.catcher_position = ( catch_area[0][0]+catch_area[1][0]/2, catch_area[0][1]+catch_area[1][1]-50 )
    self.catcher = Catcher( self.catcher_position, self.catcher_group )   
    self.game_mode = GAME_MODE_MAIN_MENU
    self.current_selected_tile = ID_MAZE_FLOOR_BGRIGHT
    self.buttons = []
    self.level_result = {}
    self.current_level = 0
    return
  
  def load_level_editor( self ):
    self.change_game_mode_to( GAME_MODE_LEVEL_EDITOR )
    return
  
  def play_level( self, level ):
    self.change_game_mode_to( GAME_MODE_PLAY_LEVEL )
    self.maze.load_level( LEVEL_FILES[level] )
    self.current_level = level
    self.ball_group = pygame.sprite.RenderPlain()
    self.level_result = {}
    self.game_paused = True
    return
  
  def play_level_1( self ):
    self.play_level( 0 )
    return
  
  def play_level_2( self ):
    self.play_level( 1 )
    return
  
  def play_level_3( self ):
    self.play_level( 2 )
    return
  
  def play_level_4( self ):
    self.play_level( 3 )
    return
  
  def play_level_5( self ):
    self.play_level( 4 )
    return
  
  def play_level_6( self ):
    self.play_level( 5 )
    return
  
  def load_main_menu( self ):
    self.change_game_mode_to( GAME_MODE_MAIN_MENU )
    return
  
  def quit_game( self ):
    self.run_game = False
    return
  
  def new_game( self ):
    # reset all level files
    for filename in LEVEL_FILES:
      f = open( filename, "r" )
      level = pickle.load( f )
      f.close()
      level.max_points_earned = 0
      f = open( filename, "w" )
      pickle.dump( level, f )
      f.close()
    # start level 1
    self.play_level( 0 )
    return
  
  def change_game_mode_to( self, new_mode ):
    if new_mode != self.game_mode:
      self.buttons = []
    self.game_mode = new_mode
    pygame.mouse.set_visible( True )
    
    if new_mode == GAME_MODE_MAIN_MENU:
      button_small_tile = get_scaled_tile(GFX_KEY_CATCH_TILES, ID_BUTTON, 150, 25)
      button_large_tile = get_scaled_tile(GFX_KEY_CATCH_TILES, ID_BUTTON, 150, 50)
      # place buttons
      button_level_1 = Button(self.play_level_1,        ((85,200),(150,30)), button_small_tile,"PLAY LEVEL 1", self.small_font)
      button_level_2 = Button(self.play_level_2,        ((315,200),(150,30)),button_small_tile,"PLAY LEVEL 2", self.small_font)
      button_level_3 = Button(self.play_level_3,        ((545,200),(150,30)),button_small_tile,"PLAY LEVEL 3", self.small_font)
      button_level_4 = Button(self.play_level_4,        ((85,395),(150,30)), button_small_tile,"PLAY LEVEL 4", self.small_font)
      button_level_5 = Button(self.play_level_5,        ((315,395),(150,30)),button_small_tile,"PLAY LEVEL 5", self.small_font)
      button_level_6 = Button(self.play_level_6,        ((545,395),(150,30)),button_small_tile,"PLAY LEVEL 6", self.small_font)
      
      # determine which levels are unlocked 
      button_level_1.enabled = True
      
      f = open( LEVEL_FILES[0], "r" )
      level = pickle.load( f )
      f.close()
      button_level_2.enabled = (level.max_points_earned >= level.point_limit)
      
      f = open( LEVEL_FILES[1], "r" )
      level = pickle.load( f )
      f.close()
      button_level_3.enabled = (level.max_points_earned >= level.point_limit)
      
      f = open( LEVEL_FILES[2], "r" )
      level = pickle.load( f )
      f.close()
      button_level_4.enabled = (level.max_points_earned >= level.point_limit)
      
      f = open( LEVEL_FILES[3], "r" )
      level = pickle.load( f )
      f.close()
      button_level_5.enabled = (level.max_points_earned >= level.point_limit)
      
      f = open( LEVEL_FILES[4], "r" )
      level = pickle.load( f )
      f.close()
      button_level_6.enabled = (level.max_points_earned >= level.point_limit)
            
      #button_level_1.enabled = True
      #button_level_2.enabled = True
      #button_level_3.enabled = True
      #button_level_4.enabled = True
      #button_level_5.enabled = True
      #button_level_6.enabled = True
      
      self.buttons.append( button_level_1 )
      self.buttons.append( button_level_2 )
      self.buttons.append( button_level_3 )
      self.buttons.append( button_level_4 )
      self.buttons.append( button_level_5 )
      self.buttons.append( button_level_6 )
      
      self.buttons.append( Button(self.new_game,          ((100,500),(150,50)),button_large_tile,"NEW GAME", self.large_font) )
      #self.buttons.append( Button(self.load_level_editor, ((300,500),(150,50)),button_large_tile,"EDITOR", self.large_font) )
      self.buttons.append( Button(self.quit_game,         ((550,500),(150,50)),button_large_tile,"QUIT", self.large_font) )
    
    if new_mode == GAME_MODE_PLAY_LEVEL:
      pygame.mouse.set_visible( False )
      pygame.mouse.set_pos( fixed_mouse_pos )
      self.maze.reset_level()
    
    if new_mode == GAME_MODE_RESULT_SCREEN:
      button_tile = get_scaled_tile(GFX_KEY_CATCH_TILES, ID_BUTTON, 200, 50)
      self.buttons.append( Button(None,   ((500,350),(200,50)),button_tile,"OK", self.large_font) )
    
    if new_mode == GAME_MODE_LEVEL_EDITOR:
      self.maze.reset_level()
      pygame.mouse.set_visible( True )
      button_tile = get_scaled_tile(GFX_KEY_CATCH_TILES, ID_BUTTON, 200, 50)
      # place buttons      
      self.buttons.append( Button(self.maze.load_level, ((600,100),(200,50)),button_tile,"LOAD", self.large_font) )
      self.buttons.append( Button(self.maze.save_level, ((600,200),(200,50)),button_tile,"SAVE", self.large_font) )
      self.buttons.append( Button(self.maze.reset_level,((600,300),(200,50)),button_tile,"RESET",self.large_font) )
      self.buttons.append( Button(self.load_main_menu,  ((600,400),(200,50)),button_tile,"MAIN MENU",self.large_font) )
      
    return
  
  def load_resources( self ):
    load_image( GFX_KEY_MASK, "game_mask.png", screen_size[0], screen_size[1] )
    load_image( GFX_KEY_CATCH_TILES, "catch_tiles.png" )
    load_image( GFX_KEY_MAZE_TILES, "maze_tiles.png" )
    load_image( GFX_KEY_MAIN_MENU, "main_menu.png" )
    load_image( GFX_KEY_BACKGROUND, "game_background.png" )
    load_image( GFX_KEY_RESULT, "result.png" )
    return
  
  def handle_play_events( self ):
    for event in pygame.event.get():
      if event.type == pygame.QUIT:
        self.run_game = False
      if event.type == pygame.MOUSEMOTION:
        new_pos = pygame.mouse.get_pos()
        if self.game_paused == False:
          rel_x = new_pos[0]-fixed_mouse_pos[0]
          rel_y = new_pos[1]-fixed_mouse_pos[1]
          
          if self.catcher.rect.left+rel_x < catch_area[0][0]:
            rel_x = self.catcher.rect.left-catch_area[0][0]
          if self.catcher.rect.right+rel_x > catch_area[0][0]+catch_area[1][0]:
            rel_x = catch_area[0][0]+catch_area[1][0] - self.catcher.rect.right
          
          if self.catcher.rect.top+rel_y < catch_area[0][1]:
            rel_y = self.catcher.rect.top-catch_area[0][1]
          if self.catcher.rect.bottom+rel_y > catch_area[0][1]+catch_area[1][1]:
            rel_y = catch_area[0][1]+catch_area[1][1] - self.catcher.rect.bottom
          
          self.catcher.rect = self.catcher.rect.move( rel_x, rel_y )
          pygame.mouse.set_pos( fixed_mouse_pos )
      
      if event.type == pygame.KEYDOWN:
        if event.key == pygame.K_ESCAPE:
          self.change_game_mode_to( GAME_MODE_MAIN_MENU )
        if event.key == pygame.K_SPACE:
          self.game_paused = not self.game_paused
          pygame.mouse.set_visible( self.game_paused )
    return
  
  def get_clicked_editor_tile_id( self, pos):
    mouse_x = pos[0]
    mouse_y = pos[1]
    tile_id = self.current_selected_tile
    
    # check if click lies inside the tile area
    if mouse_x < maze_tiles_area[0][0]:
      return tile_id
    if mouse_x > maze_tiles_area[0][0]+ID_NOF_MAZE_TILES*MAZE_TEXTURE_WIDTH:
      return tile_id
    if mouse_y < maze_tiles_area[0][1]:
      return tile_id
    if mouse_y > maze_tiles_area[0][1]+MAZE_TEXTURE_HEIGHT:
      return tile_id
    
    # calculate clicked index
    tile_id =  int( (mouse_x-maze_tiles_area[0][0])/MAZE_TEXTURE_WIDTH )
    return tile_id
  
  def handle_editor_events( self ):
    for event in pygame.event.get():
      if event.type == pygame.QUIT:
        self.run_game = False
      if event.type == pygame.KEYDOWN:
        if event.key == pygame.K_ESCAPE:
          self.change_game_mode_to( GAME_MODE_MAIN_MENU )
      if event.type == pygame.MOUSEBUTTONDOWN:
        mouse_pos = pygame.mouse.get_pos()
        # check if tile selector was clicked
        self.current_selected_tile = self.get_clicked_editor_tile_id( mouse_pos )
        # send the click position to the maze, maybe it was clicked
        self.maze.handle_click_event( mouse_pos, self.current_selected_tile )
        # check buttons
        for button in self.buttons:
          button.handle_click_event( mouse_pos ) 
    return
  
  def handle_result_screen_events( self ):
    for event in pygame.event.get():
      if event.type == pygame.QUIT:
        self.run_game = False
      if event.type == pygame.KEYDOWN:
        if event.key == pygame.K_ESCAPE:
          self.change_game_mode_to( GAME_MODE_MAIN_MENU )
      if event.type == pygame.MOUSEBUTTONDOWN:
        self.change_game_mode_to( GAME_MODE_MAIN_MENU )
    return
  
  def handle_main_menu_events( self ):
    for event in pygame.event.get():
      if event.type == pygame.QUIT:
        self.run_game = False
      if event.type == pygame.KEYDOWN:
        if event.key == pygame.K_ESCAPE:
          self.run_game = False
      if event.type == pygame.MOUSEBUTTONDOWN:
        mouse_pos = pygame.mouse.get_pos()
        # check buttons
        for button in self.buttons:
          button.handle_click_event( mouse_pos ) 
    return
  
  def handle_events( self ):
    if self.game_mode == GAME_MODE_MAIN_MENU:
      self.handle_main_menu_events()
    if self.game_mode == GAME_MODE_PLAY_LEVEL:
      self.handle_play_events()
    if self.game_mode == GAME_MODE_LEVEL_EDITOR:
      self.handle_editor_events()
    if self.game_mode == GAME_MODE_RESULT_SCREEN:
      self.handle_result_screen_events()
    return
  
  def draw_main_menu( self ):
    self.screen.blit( gfx[GFX_KEY_MAIN_MENU], (0,0) )
    # draw buttons
    for button in self.buttons:
      button.draw( self.screen )
    return
  
  def draw_result_screen( self ):
    self.draw_play_level()
    topleft = (screen_size[0]/2-result_screen_dimensions[0]/2,screen_size[1]/2-result_screen_dimensions[1]/2)
    self.screen.blit( gfx[GFX_KEY_RESULT], topleft )
    
    # print success
    draw_text_center( self.level_result[ID_FINISH_REASON] , (screen_size[0]/2-20,160), result_screen_text_color, self.large_font, self.screen)
    
    # draw points
    points = self.points_for_combo(ID_NOF_1_COMBOS) * self.level_result[ID_NOF_1_COMBOS]
    text =  str(self.level_result[ID_NOF_1_COMBOS]) +  " x 1-Combo = " + str(points) + " Points"
    draw_text_center( text , (screen_size[0]/2-20,200), result_screen_text_color, self.large_font, self.screen)
    
    points = self.points_for_combo(ID_NOF_2_COMBOS) * self.level_result[ID_NOF_2_COMBOS]
    text =  str(self.level_result[ID_NOF_2_COMBOS]) +  " x 2-Combo = " + str(points) + " Points"
    draw_text_center( text , (screen_size[0]/2-20,230), result_screen_text_color, self.large_font, self.screen)
    
    points = self.points_for_combo(ID_NOF_3_COMBOS) * self.level_result[ID_NOF_3_COMBOS]
    text =  str(self.level_result[ID_NOF_3_COMBOS]) +  " x 3-Combo = " + str(points) + " Points"
    draw_text_center( text , (screen_size[0]/2-20,260), result_screen_text_color, self.large_font, self.screen)
    
    points = self.points_for_combo(ID_NOF_4_COMBOS) * self.level_result[ID_NOF_4_COMBOS]
    text =  str(self.level_result[ID_NOF_4_COMBOS]) +  " x 4-Combo = " + str(points) + " Points"
    draw_text_center( text , (screen_size[0]/2-20,290), result_screen_text_color, self.large_font, self.screen)
    
    points = self.points_for_combo(ID_NOF_5_COMBOS) * self.level_result[ID_NOF_5_COMBOS]
    text =  str(self.level_result[ID_NOF_5_COMBOS]) +  " x 5-Combo = " + str(points) + " Points"
    draw_text_center( text , (screen_size[0]/2-20,320), result_screen_text_color, self.large_font, self.screen)
    
    points = self.points_for_combo(ID_NOF_6_COMBOS) * self.level_result[ID_NOF_6_COMBOS]
    text =  str(self.level_result[ID_NOF_6_COMBOS]) +  " x 6-Combo = " + str(points) + " Points"
    draw_text_center( text , (screen_size[0]/2-20,350), result_screen_text_color, self.large_font, self.screen)
    
    text =  "<Mouseclick> to continue..."
    draw_text_center( text , (screen_size[0]/2-20,400), result_screen_text_color, self.large_font, self.screen)
    
    return
  
  def draw_play_level( self ):
    # draw catcher stuff
    self.containers.balls.draw( self.screen )
    self.ball_group.draw( self.screen )
    self.catcher_group.draw( self.screen )
    
    # draw maze stuff
    self.maze.draw( self.screen )
            
    self.screen.blit( gfx[GFX_KEY_MASK], (0,0) )
    
    # print time limit
    time_string = "Time:" + str(int(self.maze.time_left)) + " seconds"
    draw_text_center( time_string , (150,30), (255,255,255), self.large_font, self.screen)
    
    points_string = str(self.level_result[ID_NOF_POINTS]) + "/" + str(self.maze.level.point_limit) + " Points"
    draw_text_center( points_string , (450,30), (255,255,255), self.large_font, self.screen)
    
    if self.game_paused:
      topx = screen_size[0]/2-result_screen_dimensions[0]/2
      topy = screen_size[1]/2-result_screen_dimensions[1]/2
      topleft = (topx,topy)
      self.screen.blit( gfx[GFX_KEY_RESULT], topleft )
      
      # show hints
      draw_text_center( "Hints", (screen_size[0]/2-5,topy+40), result_screen_text_color, self.small_font, self.screen )
      for hint in self.maze.level.hints:
        draw_text_center( hint, (screen_size[0]/2-5,topy+90+30*self.maze.level.hints.index(hint)), result_screen_text_color, self.small_font, self.screen )
    return
  
  def draw_level_editor( self ):
    # draw maze tiles
    for index in range(ID_NOF_MAZE_TILES):
      surface = get_scaled_tile( GFX_KEY_MAZE_TILES, index, MAZE_TEXTURE_WIDTH, MAZE_TEXTURE_HEIGHT )
      x = maze_tiles_area[0][0]+index*MAZE_TEXTURE_WIDTH
      y = maze_tiles_area[0][1]
      self.screen.blit( surface, (x,y) )
      # highlight current selected tile
      if index == self.current_selected_tile:
        pygame.draw.rect( self.screen, highlight_selected_tile_color, pygame.Rect(x, y, MAZE_TEXTURE_WIDTH, MAZE_TEXTURE_HEIGHT), 1)
    
    # draw maze stuff
    self.maze.draw( self.screen, GAME_MODE_LEVEL_EDITOR )
    
    # draw buttons
    for button in self.buttons:
      button.draw( self.screen )

    return
  
  def draw( self ):
    self.screen.fill( background_color )
    
    if self.game_mode == GAME_MODE_MAIN_MENU:
      self.draw_main_menu()
    
    if self.game_mode == GAME_MODE_PLAY_LEVEL:
      self.draw_play_level()
      
    if self.game_mode == GAME_MODE_LEVEL_EDITOR:
      self.draw_level_editor()
    
    if self.game_mode == GAME_MODE_RESULT_SCREEN:
      self.draw_result_screen()
    
    pygame.display.flip()
    return
  
  def tick_play_level( self ):
    # check if there appears a new ball
    if self.game_paused == False:
      if random.randrange(new_ball_probability) == 0:
        position = get_new_command_ball_position( self.ball_group )
        if position != (0,0):
          CommandBall( position, self.ball_group, random.randrange( self.maze.level.nof_balls) )
      
    # update objects
    if self.game_paused == False:
      for ball in self.ball_group:
        ball.tick( self.maze.level.ball_speed )
      self.maze.robot_tick( self.containers )
      
      # collision detection
      balls = pygame.sprite.spritecollide( self.catcher, self.ball_group, False )
      for ball in balls:
        self.ball_group.remove( ball )
        self.containers.add_ball( ball )
    # update points
    self.calculate_level_results()
    return
  
  def tick( self ):
    if self.game_mode == GAME_MODE_PLAY_LEVEL:
      self.tick_play_level()
    return
  
  def points_for_combo( self, combo ):
    if combo == ID_NOF_1_COMBOS:
      return 1
    if combo == ID_NOF_2_COMBOS:
      return 3
    if combo == ID_NOF_3_COMBOS:
      return 6
    if combo == ID_NOF_4_COMBOS:
      return 10
    if combo == ID_NOF_5_COMBOS:
      return 15
    if combo == ID_NOF_6_COMBOS:
      return 21
    return 0
  
  def add_combo_points( self, combo):
    if combo == ID_NOF_1_COMBOS:
      self.level_result[ID_NOF_1_COMBOS] = self.level_result[ID_NOF_1_COMBOS] + 1
    if combo == ID_NOF_2_COMBOS:
      self.level_result[ID_NOF_2_COMBOS] = self.level_result[ID_NOF_2_COMBOS] + 1
    if combo == ID_NOF_3_COMBOS:
      self.level_result[ID_NOF_3_COMBOS] = self.level_result[ID_NOF_3_COMBOS] + 1
    if combo == ID_NOF_4_COMBOS:
      self.level_result[ID_NOF_4_COMBOS] = self.level_result[ID_NOF_4_COMBOS] + 1
    if combo == ID_NOF_5_COMBOS:
      self.level_result[ID_NOF_5_COMBOS] = self.level_result[ID_NOF_5_COMBOS] + 1
    if combo == ID_NOF_6_COMBOS:
      self.level_result[ID_NOF_6_COMBOS] = self.level_result[ID_NOF_6_COMBOS] + 1
    return
  
  def calculate_level_results( self ):
    # clear old results
    self.level_result = {}
    self.level_result[ID_NOF_1_COMBOS] = 0
    self.level_result[ID_NOF_2_COMBOS] = 0
    self.level_result[ID_NOF_3_COMBOS] = 0
    self.level_result[ID_NOF_4_COMBOS] = 0
    self.level_result[ID_NOF_5_COMBOS] = 0
    self.level_result[ID_NOF_6_COMBOS] = 0
    self.level_result[ID_NOF_POINTS] = 0
    
    self.level_result[ID_FINISH_REASON] = ""
    # count points
    last_type = -1
    nof_equal_balls = 0
    
    for ball in self.containers.robotq:
      current_type = ball.type
      if last_type == -1:
        last_type = current_type
        nof_equal_balls = 1
        continue
      else:
        if last_type == current_type:
          nof_equal_balls = nof_equal_balls + 1
          continue
        else:
          self.add_combo_points( nof_equal_balls )
          if current_type == ID_CMD_BALL_FINISH:
            last_type = -1
            nof_equal_balls = 0
            continue
          else:
            last_type = current_type
            nof_equal_balls = 1
            continue
    self.add_combo_points( nof_equal_balls )
    self.level_result[ID_NOF_POINTS] = self.level_result[ID_NOF_POINTS] + self.points_for_combo(ID_NOF_1_COMBOS)*self.level_result[ID_NOF_1_COMBOS]
    self.level_result[ID_NOF_POINTS] = self.level_result[ID_NOF_POINTS] + self.points_for_combo(ID_NOF_2_COMBOS)*self.level_result[ID_NOF_2_COMBOS]
    self.level_result[ID_NOF_POINTS] = self.level_result[ID_NOF_POINTS] + self.points_for_combo(ID_NOF_3_COMBOS)*self.level_result[ID_NOF_3_COMBOS]
    self.level_result[ID_NOF_POINTS] = self.level_result[ID_NOF_POINTS] + self.points_for_combo(ID_NOF_4_COMBOS)*self.level_result[ID_NOF_4_COMBOS]
    self.level_result[ID_NOF_POINTS] = self.level_result[ID_NOF_POINTS] + self.points_for_combo(ID_NOF_5_COMBOS)*self.level_result[ID_NOF_5_COMBOS]
    self.level_result[ID_NOF_POINTS] = self.level_result[ID_NOF_POINTS] + self.points_for_combo(ID_NOF_6_COMBOS)*self.level_result[ID_NOF_6_COMBOS]
    
    if self.level_result[ID_NOF_POINTS] >= self.maze.level.point_limit:
      self.level_result[ID_FINISH_REASON] = "Level cleared!"
    else:
      self.level_result[ID_FINISH_REASON] = "Not enough points!"
    return
  
  def check_win_or_loose_conditions( self ):
    if self.game_mode != GAME_MODE_PLAY_LEVEL:
      return
    if self.maze.robot_position == self.maze.level.robot_exit:
      #self.change_game_mode_to( GAME_MODE_MAIN_MENU )
      self.calculate_level_results()
      
      # save result in level file
      self.maze.level.max_points_earned = max( self.maze.level.max_points_earned, self.level_result[ID_NOF_POINTS] )
      f = open( LEVEL_FILES[self.current_level], "w" )
      pickle.dump( self.maze.level, f )
      f.close()
      
      self.change_game_mode_to( GAME_MODE_RESULT_SCREEN )
    if self.maze.time_left < 0:
      self.calculate_level_results()
      self.level_result[ID_FINISH_REASON] = "Time out!"
      self.change_game_mode_to( GAME_MODE_RESULT_SCREEN )
    return
  
  def start_game_loop( self ):
    while self.run_game:
      # handle events
      self.handle_events()
      # update objects
      self.tick()
      # draw stuff
      self.draw()
      # check wining condition
      self.check_win_or_loose_conditions()
      
      # wait a little to get 30fps
      pygame.time.delay( 1000/int(frames_per_second) )
    return

def main():
  game = Game()
  game.change_game_mode_to( GAME_MODE_MAIN_MENU )
  game.start_game_loop()
  pygame.quit()